--
-- Busy Bee Tasks plugin for Lord of The Rings Online
--
-- by Bosko
--
-- http://www.lotrointerface.com/downloads/fileinfo.php?id=536
--

import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Extensions";
import "Turbine.UI.Lotro";
import "BoskoPlugins.BusyBee.BusyBeeTasks"

function writeln(str)
    Turbine.Shell.WriteLine(str);
end

--==============================================================================
-- Settings
--==============================================================================

BusyBeeSettingsVersion = 1.1;

local HeadingTextColor = Turbine.UI.Color(0.80, 0.80, 0.30);
local EntryTextColor   = Turbine.UI.Color(1.00, 1.00, 0.70);
local ItemLabelColor   = Turbine.UI.Color(0.50, 0.50, 0.50);

local clrLevelTooLow0  = Turbine.UI.Color(0.18, 0.18, 0.18);
local clrLevelTooLow1  = Turbine.UI.Color(0.28, 0.28, 0.28);
local clrLevelTooLow2  = Turbine.UI.Color(0.31, 0.35, 0.32);
local clrLevelInRange0 = Turbine.UI.Color(0.46, 0.41, 0.25);
local clrLevelInRange1 = Turbine.UI.Color(0.80, 0.80, 0.47);
local clrLevelInRange2 = Turbine.UI.Color(0.33, 0.80, 0.15);
local clrLevelTooHigh0 = Turbine.UI.Color(0.35, 0.21, 0.21);
local clrLevelTooHigh1 = Turbine.UI.Color(0.50, 0.30, 0.30);
local clrLevelTooHigh2 = Turbine.UI.Color(0.70, 0.42, 0.35);

local headingWidth = 400;
local headingHeight = 16;
local entryHeight = 14;

local windowOpacityActive = 1.0;
local windowOpacityPassive = 0.9;
local clrBackgroundActive = Turbine.UI.Color(0.5,0.0,0.0,0.0);
local clrBackgroundPassive = Turbine.UI.Color(0.0,0.0,0.0,0.0);

local windowOriginalWidth = 415;
local windowOriginalHeight = 200;

local marginX = 10;
local marginTop = 5;
local marginBottom = 10;
local buttonWidth1 = 60;
local buttonWidth2 = 40;
local buttonHeight = 20;

--==============================================================================
-- Data Generation
--==============================================================================

--------------------------------------------------------------------------------
local function BusyBeeSortById(itemA, itemB)

    return itemA.id < itemB.id;
end

--------------------------------------------------------------------------------
local function BusyBeeSortByLocation(itemA, itemB)

    if itemA.location == itemB.location then
        return BusyBeeSortById(itemA, itemB);
    else
        return itemA.location < itemB.location;
    end
end

--------------------------------------------------------------------------------
local function BusyBeeSortByFullName(itemA, itemB)

    if itemA.fullname == itemB.fullname then
        return BusyBeeSortByLocation(itemA, itemB);
    else
        return itemA.fullname < itemB.fullname;
    end
end

--------------------------------------------------------------------------------
local function BusyBeeSortByLevel(itemA, itemB)

    if itemA.lvlmin == itemB.lvlmin then
        return BusyBeeSortByFullName(itemA, itemB);
    else
        return itemA.lvlmin < itemB.lvlmin;
    end
end

--------------------------------------------------------------------------------
local function BusyBeeSortByName(itemA, itemB)

    if itemA.name == itemB.name then
        return BusyBeeSortByLevel(itemA, itemB);
    else
        return itemA.name < itemB.name;
    end
end

--------------------------------------------------------------------------------
local function BusyBeeSortByFaction(itemA, itemB)

    if itemA.faction == itemB.faction then
        return BusyBeeSortByLevel(itemA, itemB);
    else
        return itemA.faction < itemB.faction;
    end
end

--------------------------------------------------------------------------------
local function BusyBeeItemName(item)

    if type(item) == "number" then
        return BusyBeeTasks[item].fullname;
    else
        return item.fullname;
    end
end

--------------------------------------------------------------------------------
-- Sort by item name
table.sort(BusyBeeTasks, BusyBeeSortByFullName);

for i, item in ipairs(BusyBeeTasks) do
    item.visible = true;
    item.nodes = { };
    item.location = item.zone .. " - " .. item.place;
    item.fullname = item.prefix .. " " .. item.name;
end

-- Build BusyBeeCounts
BusyBeeCounts = { };
for i, item in ipairs(BusyBeeTasks) do
    local index = BusyBeeItemName(item);
    local tableItem = BusyBeeCounts[index];
    if (tableItem == nil) then
        -- Initialize the item
        BusyBeeCounts[index] = {count = 0, quantity = item.quantity, ids = {}, nodes = {}};
    end
    
    -- Add item information
    table.insert(BusyBeeCounts[index].ids, item.id);
end

-- Sort by name
table.sort(BusyBeeTasks, BusyBeeSortByLevel);

-- Build BusyBeeLevels
BusyBeeLevels = { };
    table.insert(BusyBeeLevels, {key = "All Items", ids = {}});
BusyBeeLevelsCount = 1;
for i, item in ipairs(BusyBeeTasks) do
    -- Add item information
    table.insert(BusyBeeLevels[BusyBeeLevelsCount].ids, item.id);
end

-- -- Sort by name
-- table.sort(BusyBeeTasks, BusyBeeSortByName);
-- 
-- -- Build BusyBeeNames
-- BusyBeeNames = { };
-- BusyBeeNamesCount = 0;
-- local lastName = "";
-- for i, item in ipairs(BusyBeeTasks) do
--     local thisName = item.name;
--     if thisName ~= lastName then
--         table.insert(BusyBeeNames, {key = thisName, ids = {}});
--         BusyBeeNamesCount = BusyBeeNamesCount + 1;
--         lastName = thisName;
--     end
--     
--     -- Add item information
--     table.insert(BusyBeeNames[BusyBeeNamesCount].ids, item.id);
-- end

-- Sort by location
table.sort(BusyBeeTasks, BusyBeeSortByLocation);

-- Build BusyBeeLocations
BusyBeeLocations = { };
BusyBeeLocationsCount = 0;
local lastLocation = "";
for i, item in ipairs(BusyBeeTasks) do
    local thisLocation = item.location;
    if thisLocation ~= lastLocation then
        table.insert(BusyBeeLocations, {key = thisLocation, ids = {}});
        BusyBeeLocationsCount = BusyBeeLocationsCount + 1;
        lastLocation = thisLocation;
    end
    
    -- Add item information
    table.insert(BusyBeeLocations[BusyBeeLocationsCount].ids, item.id);
end

-- Sort by faction
table.sort(BusyBeeTasks, BusyBeeSortByFaction);

-- Build BusyBeeFactions
BusyBeeFactions = { };
BusyBeeFactionsCount = 0;
local lastFaction = "";
for i, item in ipairs(BusyBeeTasks) do
    local thisFaction = item.faction;
    if thisFaction ~= lastFaction then
        table.insert(BusyBeeFactions, {key = thisFaction, ids = {}});
        BusyBeeFactionsCount = BusyBeeFactionsCount + 1;
        lastFaction = thisFaction;
    end
    
    -- Add item information
    table.insert(BusyBeeFactions[BusyBeeFactionsCount].ids, item.id);
end

-- Sort by id
table.sort(BusyBeeTasks, BusyBeeSortById);

--==============================================================================
-- Tree Data
--==============================================================================

local BusyBeeTreeData = {
    { data = BusyBeeLocations, fields = {
        { key = "level",  width = 50, alignment = "left" }, 
        { key = "fullname",  width = 110, alignment = "left" }, 
        { key = "count",  width = 60, alignment = "center" }, 
        { key = "faction",  width = 165, alignment = "left" }, 
        { key = "sellvalue",  width = 50, alignment = "center" }
    } },
    { data = BusyBeeFactions, fields = {
        { key = "level",  width = 50, alignment = "left" }, 
        { key = "fullname",  width = 110, alignment = "left" }, 
        { key = "count",  width = 60, alignment = "center" }, 
        { key = "location",  width = 165, alignment = "left" }, 
        { key = "sellvalue",  width = 50, alignment = "center" }, 
    } },
--     { data = BusyBeeNames, fields = {
--         { key = "level",  width = 50, alignment = "left" }, 
--         { key = "fullname",  width = 115, alignment = "left" }, 
--         { key = "count",  width = 55, alignment = "center" }, 
--         { key = "location",  width = 165, alignment = "left" }, 
--         { key = "faction",  width = 165, alignment = "left" }, 
--         { key = "sellvalue",  width = 50, alignment = "center" }, 
--     } },
    { data = BusyBeeLevels, fields = {
        { key = "level",  width = 50, alignment = "left" }, 
        { key = "fullname",  width = 110, alignment = "left" }, 
        { key = "count",  width = 60, alignment = "center" }, 
        { key = "faction",  width = 125, alignment = "left" }, 
        { key = "location",  width = 165, alignment = "left" }, 
        { key = "sellvalue",  width = 50, alignment = "center" }, 
    } },
};


--==============================================================================
-- BusyBeeLabel
--==============================================================================
BusyBeeLabel = class(Turbine.UI.Label);

function BusyBeeLabel:Constructor(parent, x, w, h, alignment)
    Turbine.UI.Label.Constructor(self);
    
    -- Set the format
    self:SetParent(parent);
    self:SetPosition(x, 0);
    self:SetSize(w, h);
    self:SetFontStyle(Turbine.UI.FontStyle.Outline);
    self:SetTextAlignment(alignment);
    self:SetForeColor(EntryTextColor);
    self:SetFont(Turbine.UI.Lotro.Font.Verdana12);
end

--==============================================================================
-- BusyBeeHeading
--==============================================================================

BusyBeeHeading = class(Turbine.UI.TreeNode);

--------------------------------------------------------------------------------
function BusyBeeHeading:Constructor(szText)
    Turbine.UI.TreeNode.Constructor(self);
    
    self:SetSize(headingWidth, headingHeight);
    
    -- Create and format the heading label
    self.header = Turbine.UI.Label();
    self.header:SetParent(self);
    self.header:SetPosition(0, 0);
    self.header:SetSize(headingWidth, headingHeight);
    self.header:SetFontStyle(Turbine.UI.FontStyle.Outline);
    self.header:SetForeColor(HeadingTextColor);
    self.header:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
    self.header:SetFont(Turbine.UI.Lotro.Font.TrajanPro15);
    self.header:SetText(szText);
end

--==============================================================================
-- BusyBeeNode
--==============================================================================

BusyBeeNode = class(Turbine.UI.TreeNode);

--------------------------------------------------------------------------------
function BusyBeeNode:Constructor(item, left, fields)
    Turbine.UI.TreeNode.Constructor(self);
    
    self.item = item;
    self.fields = fields;
    self.countField = 0;
    
    -- Full name and quantity are fixed values
    self.quantity = self:GetField(item, "quantity");
    self.fullname = self:GetField(item, "fullname");
    
    -- Create and populate the labels
    local nFields = 0;
    for i, desc in ipairs(fields) do
        nFields = nFields + 1;
        
        -- Identify alignment
        alignment = Turbine.UI.ContentAlignment.MiddleCenter;
        if (desc.alignment == "left") then
            alignment = Turbine.UI.ContentAlignment.MiddleLeft;
        elseif (desc.alignment == "right") then
            alignment = Turbine.UI.ContentAlignment.MiddleRight;
        end
        
        -- Create the label and fill the text
        self["field" .. nFields] = BusyBeeLabel(self, left, desc.width, entryHeight, alignment);
        self["field" .. nFields]:SetText(self:FormatText(desc.key));
        left = left + desc.width;
        
        -- Remember which field is "count"
        if desc.key == "count" then
            self.countField = nFields;
        end
    end
    
    -- Set the size and update the count
    self:SetSize(left, entryHeight);
    self:UpdateCount();
end

--------------------------------------------------------------------------------
function BusyBeeNode:GetField(item, field)
    local value = item[field];
    if (value == nil) then
        local n = 0;
        for i, value in ipairs(item.ids) do
            local item2 = BusyBeeTasks[id];
            if item2.visible then
                n = n + 1;
                if (n == 1) then
                    -- First field
                    value = item2[field];
                elseif n == 2 then
                    if value ~= item2[field] then
                        -- Different field, mark as such
                        value = "multiple " .. field .. "s";
                    else
                        -- Same field as before
                        n = 1;
                    end
                end
            end
        end
    end
    
    if value == nil then
        value = "???";
    end
    return value;
end

--------------------------------------------------------------------------------
function BusyBeeNode:GetFieldRange(item, field)
    local minValue = item[field];
    local maxValue = item[field];
    if (minValue == nil) then
        minValue = 0;
        maxValue = 0;
        local n = 0;
        for i, value in ipairs(item.ids) do
            local item2 = BusyBeeTasks[id];
            if item2.visible then
                n = n + 1;
                if (n == 1) then
                    minValue = item2[field];
                    maxValue = item2[field];
                else
                    if (minValue > item2[field]) then
                        minValue = item2[field];
                    elseif (maxValue < item2[field]) then
                        maxValue = item2[field];
                    end
                end
            end
        end
    end
    return minValue, maxValue;
end


--------------------------------------------------------------------------------
function BusyBeeNode:FormatText(field)

    local id = self.item.id;
    if id == nil then
        if type(self.item.ids) ~= "table" then
            return "";
        end
        id = self.item.ids[1];
    end
    
    if (field == "level") then
        local min1, max1 = self:GetFieldRange(self.item, "lvlmin");
        local min2, max2 = self:GetFieldRange(self.item, "lvlmax");
        return string.format("[%d-%d]", min1, max2);
    elseif (field == "fullname") then
        return BusyBeeItemName(id);
    elseif (field == "count") then
        return "";--string.format("%d/%d", BusyBeeCounts[fullname].count, BusyBeeCounts[fullname].quantity);
    elseif (field == "sellvalue") then
        return string.format("%.2f", BusyBeeTasks[id].sellvalue);
    else
        return self:GetField(self.item, field);
    end
end

--------------------------------------------------------------------------------
function BusyBeeNode:UpdateCount()
    local player = Turbine.Gameplay.LocalPlayer.GetInstance();
    local playerLevel = player:GetLevel();
    local lvlMin, lvlMax2 = self:GetFieldRange(self.item, "lvlmin");
    local lvlMin2, lvlMax = self:GetFieldRange(self.item, "lvlmax");
    id = self.item.id;
    if id == nil then
        if type(self.item.ids) ~= "table" then
            return "";
        end
        id = self.item.ids[1];
    end
    
    local count = BusyBeeCounts[self.fullname].count;
    local quantity = BusyBeeCounts[self.fullname].quantity;
    
    local clr = clrLevelInRange0;
    
    self["field" .. self.countField]:SetText(string.format("%d/%d", count, quantity));
    
    if playerLevel > lvlMax then
        if count == 0 then
            clr = clrLevelTooLow0;
        elseif count < quantity then
            clr = clrLevelTooLow1;
        else
            clr = clrLevelTooLow2;
        end
    elseif playerLevel < lvlMin then
        if count == 0 then
            clr = clrLevelTooHigh0;
        elseif count < quantity then
            clr = clrLevelTooHigh1;
        else
            clr = clrLevelTooHigh2;
        end
    else
        if count == 0 then
            clr = clrLevelInRange0;
        elseif count < quantity then
            clr = clrLevelInRange1;
        else
            clr = clrLevelInRange2;
        end
    end
    
    self.field1:SetForeColor(clr);
    self.field2:SetForeColor(clr);
    self.field3:SetForeColor(clr);
    self.field4:SetForeColor(clr);
    self.field5:SetForeColor(clr);
end

--------------------------------------------------------------------------------
function BusyBeeNode:FixVisibility()
    local oldHeight = self:GetHeight();
    local height = 0;
    
    for i, desc in ipairs(self.fields) do
        if i ~= self.countField then
            self["field" .. i]:SetText(self:FormatText(desc.key));
        end
    end
    
    if self.item.id ~= nil then
        if BusyBeeTasks[self.item.id].visible then
            height = entryHeight;
        end
    else
        for i, id in ipairs(self.item.ids) do
            if BusyBeeTasks[id].visible then
                height = entryHeight;
            end
        end
    end
    
    if oldHeight ~= height then
        self:SetHeight(height);
        
        local parentNode = self:GetParentNode();
        local childNodes = parentNode:GetChildNodes();
        local fFound = 0;
        for j = 1,childNodes:GetCount(), 1 do
            local subItem = childNodes:Get(j);
            if 0 < subItem:GetHeight() then
                fFound = 1;
                break;
            end
        end
        parentNode:SetHeight(headingHeight * fFound);
    end
end

--==============================================================================
-- BusyBeeTree
--==============================================================================
BusyBeeTree = class(Turbine.UI.Control);

--------------------------------------------------------------------------------
function BusyBeeTree:Constructor(items, fields)
    Turbine.UI.Control.Constructor(self);
    
    -- Create tree view
    self.treeView = Turbine.UI.TreeView();
    self.treeView:SetParent(self);

    -- Create tree view scrollbar
    self.scrollBar = Turbine.UI.Lotro.ScrollBar();
    self.scrollBar:SetOrientation(Turbine.UI.Orientation.Vertical);
    self.scrollBar:SetParent(self);
    self.treeView:SetVerticalScrollBar(self.scrollBar);
    self.showScrollbar = true;

    -- Populate the tree view
    local level1 = self.treeView:GetNodes();
    
    for i, value in ipairs(items) do
        local heading = BusyBeeHeading(value.key);
        level1:Add(heading);
        local level2 = heading:GetChildNodes();
        for i, id in ipairs(value.ids) do
            local item = BusyBeeTasks[id];
            local node = BusyBeeNode(item, 0, fields);
            table.insert(BusyBeeCounts[BusyBeeItemName(id)].nodes, node);
            table.insert(BusyBeeTasks[id].nodes, node);
            level2:Add(node);
        end
    end
    
    self.SizeChanged = function(sender, args)
        sender:FixLayout();
    end
end

--------------------------------------------------------------------------------
function BusyBeeTree:GetState()

    -- Compute state
    local level1 = self.treeView:GetNodes();
    local treeState = 0;
    for i = 1,level1:GetCount(), 1 do
        local item = level1:Get(i);
        local level2 = item:GetChildNodes();
        local subItem = level2:Get(1);
        treeState = 2 * treeState;
        if subItem:IsVisible() then
            treeState = treeState + 1;
        end
    end
    
    return treeState;
end

--------------------------------------------------------------------------------
function BusyBeeTree:RestoreState(treeState)

    -- Restore tree state (expand appropriate nodes)
    local level1 = self.treeView:GetNodes();
    for i = level1:GetCount(), 1, -1 do
        if (treeState % 2 ~= 0) then
            local item = level1:Get(i);
            item:SetExpanded(true);
            treeState = treeState - 1;
        end
        treeState = treeState / 2;
    end
end

--------------------------------------------------------------------------------
function BusyBeeTree:FixLayout()
    self:Layout({ });
end

--------------------------------------------------------------------------------
function BusyBeeTree:Layout(args)
    local left, top = self:GetPosition();
    local width, height = self:GetSize();
    local scrollWidth = 10;
    
    self.scrollBar:SetPosition(width - scrollWidth, 0);
    self.scrollBar:SetSize(scrollWidth, height);
    self.treeView:SetPosition(0, 0);
    self.treeView:SetSize(width - scrollWidth, height);
end

-- --------------------------------------------------------------------------------
-- function BusyBeeTree:FixVisibility()
-- 
--     -- Hide zones
--     local level1 = self.treeView:GetNodes();
--     for i = 1,level1:GetCount(), 1 do
--         local item = level1:Get(i);
--         local level2 = item:GetChildNodes();
--         local fFound = 0;
--         for j = 1,level2:GetCount(), 1 do
--             local subItem = level2:Get(j);
--             if 0 < subItem:GetHeight() then
--                 fFound = 1;
--                 break;
--             end
--         end
--         item:SetHeight(headingHeight * fFound);
--     end
-- end

--------------------------------------------------------------------------------
function BusyBeeTree:ExpandAll()
    self.treeView:ExpandAll();
end

--------------------------------------------------------------------------------
function BusyBeeTree:CollapseAll()
    self.treeView:CollapseAll();
end

--------------------------------------------------------------------------------
function BusyBeeTree:ShowScrollBar(show)
    self.showScrollbar = show;
    self.scrollBar:SetVisible(show);
end

--==============================================================================
-- BusyBeeWindow
--==============================================================================
BusyBeeWindow = class(Turbine.UI.Extensions.SimpleWindow);

--------------------------------------------------------------------------------
function BusyBeeWindow:Constructor()
    Turbine.UI.Extensions.SimpleWindow.Constructor(self);

    self:LoadSettings();

    self.dragging = false;
    
    -- Position the window in top-center
    self:SetSize(self.settings.windowWidth, self.settings.windowHeight);
    self:SetBackColor(clrBackgroundPassive);
    self:SetPosition(self.settings.windowLeft, self.settings.windowTop);
    self:SetText("Busy Bee Tasks");
    self:SetMouseVisible(true);

    -- Set up mouse-triggered fade
    self:SetFadeSpeed(0.4);
    self:SetPassiveOpacity(self.settings.passiveOpacity);

    self.MouseEnter = function(sender, args)
        if not sender.dragging then
            sender:SetOpacity(windowOpacityActive);
            sender.fShowButtons = true;
            sender:SetBackColor(clrBackgroundActive);
            for i, tree in ipairs(sender.trees) do
                tree:ShowScrollBar(true);
            end
            sender:PerformLayout();
        end
    end

    self.MouseLeave = function(sender, args)
        if not sender.dragging then
            sender:SetOpacity(windowOpacityPassive);
            sender.fShowButtons = false;
            sender:SetBackColor(clrBackgroundPassive);
            for i, tree in ipairs(sender.trees) do
                tree:ShowScrollBar(false);
            end
            sender:PerformLayout();
            sender:SaveSettings();
        end
    end

    self.player = Turbine.Gameplay.LocalPlayer.GetInstance();
    self.backpack = self.player:GetBackpack();

    local mainWindow = self;
    
    -- Level change handler
     self.player.LevelChanged = function(sender, args)
        for i, item in ipairs(BusyBeeTasks) do
            for j, node in ipairs(item.nodes) do
                node:UpdateCount();
            end
        end
        self:UpdateAllTasks();
    end

    -- Button-related states
    self.fShowButtons = false;
    self.fShowZero = self.settings.fShowZero;
    self.fShowAllLevels = self.settings.fShowAllLevels;
    self.fShowRepAward = self.settings.fShowRepAward;
    
    -- Create mode button
    self.ModeButton = Turbine.UI.Lotro.Button();
    self.ModeButton:SetParent(self);
    self.ModeButton:SetFont(Turbine.UI.Lotro.Font.TrajanPro13);
    self.ModeButton:SetText("Next");
    self.ModeButton.Click = function(sender, args)
        local fNextTree = false;
        for i, tree in ipairs(mainWindow.trees) do
            if fNextTree then
                tree:SetVisible(true);
                fNextTree = false;
            elseif tree:IsVisible() then
                tree:SetVisible(false);
                fNextTree = true;
            end
        end
        if fNextTree then
            mainWindow.trees[1]:SetVisible(true);
        end
    end
    
    -- Create item count toggle button
    self.ItemCountButton = Turbine.UI.Lotro.GoldButton();
    self.ItemCountButton:SetParent(self);
    self.ItemCountButton:SetFont(Turbine.UI.Lotro.Font.TrajanPro13);
    self.ItemCountButton.Click = function(sender, args)
        mainWindow.fShowZero = not mainWindow.fShowZero;
        mainWindow:UpdateAllTasks();
        mainWindow:PerformLayout();
    end
    
    -- Create level eligibility toggle button
    self.AllLevelsButton = Turbine.UI.Lotro.GoldButton();
    self.AllLevelsButton:SetParent(self);
    self.AllLevelsButton:SetFont(Turbine.UI.Lotro.Font.TrajanPro13);
    self.AllLevelsButton.Click = function(sender, args)
        mainWindow.fShowAllLevels = not mainWindow.fShowAllLevels;
        mainWindow:UpdateAllTasks();
        mainWindow:PerformLayout();
    end
    
    -- Create reputation award toggle button
    self.RepAwardButton = Turbine.UI.Lotro.GoldButton();
    self.RepAwardButton:SetParent(self);
    self.RepAwardButton:SetFont(Turbine.UI.Lotro.Font.TrajanPro13);
    self.RepAwardButton.Click = function(sender, args)
        mainWindow.fShowRepAward = not mainWindow.fShowRepAward;
        mainWindow:UpdateAllTasks();
        mainWindow:PerformLayout();
    end
    
    -- Create 'Expand All' button
    self.ExpandButton = Turbine.UI.Lotro.Button();
    self.ExpandButton:SetParent(self);
    self.ExpandButton:SetFont(Turbine.UI.Lotro.Font.TrajanPro17);
    self.ExpandButton:SetText("+");
    self.ExpandButton.Click = function(sender, args)
        for i, tree in ipairs(self.trees) do
            if tree:IsVisible() then
                tree:ExpandAll();
            end
        end
    end
    
    -- Create 'Collapse All' button
    self.CollapseButton = Turbine.UI.Lotro.Button();
    self.CollapseButton:SetParent(self);
    self.CollapseButton:SetFont(Turbine.UI.Lotro.Font.TrajanPro17);
    self.CollapseButton:SetText("-");
    self.CollapseButton.Click = function(sender, args)
        for i, tree in ipairs(self.trees) do
            if tree:IsVisible() then
                tree:CollapseAll();
            end
        end
    end

    -- Create the resize handle
    self.resizeHandle = Turbine.UI.Control();
    self.resizeHandle:SetParent(self);
    self.resizeHandle:SetZOrder(100);
    self.resizeHandle:SetSize(20, 20);
    self.resizeHandle:SetPosition(self:GetWidth() - self.resizeHandle:GetWidth(), 
        self:GetHeight() - self.resizeHandle:GetHeight());

    self.resizeHandle.MouseDown = function(sender, args)
        sender.dragStartX = args.X;
        sender.dragStartY = args.Y;
        sender.dragging = true;
    end

    self.resizeHandle.MouseMove = function(sender, args)
        local width, height = mainWindow:GetSize();

        if (sender.dragging) then
            mainWindow:SetSize(width + (args.X - sender.dragStartX), height + (args.Y - sender.dragStartY));
            sender:SetPosition(mainWindow:GetWidth() - sender:GetWidth(), mainWindow:GetHeight() - sender:GetHeight())
            mainWindow:PerformLayout();
        end
    end

    self.resizeHandle.MouseUp = function(sender, args)
        sender.dragging = false;
    end

    -- Setup the window move behavior
    self.MouseDown = function(sender, args)
        sender.dragStartX = args.X;
        sender.dragStartY = args.Y;
        sender.dragging = true;
    end

    self.MouseMove = function(sender, args)
        local width, height = mainWindow:GetSize();

        if (sender.dragging) then
            sender:SetPosition(mainWindow:GetLeft() + (args.X - sender.dragStartX), 
                mainWindow:GetTop() + (args.Y - sender.dragStartY));
            mainWindow:PerformLayout();
        end
    end

    self.MouseUp = function(sender, args)
        sender.dragging = false;
    end

    -- Set up QuantityChanged callbacks for all items
    local backpackSize = self.backpack:GetSize();
    for j = 1, backpackSize, 1 do
        local item = self.backpack:GetItem(j);
        if item and item.QuantityChanged == nil then
            item.QuantityChanged = function(sender, args)
                if item:GetCategory() == 164 then
                    mainWindow:UpdateCount(item:GetName());
                end
            end
        end
    end

    -- Backpack size changed should almost never happen
    self.backpack.SizeChanged = function(sender, args)
        -- Set up QuantityChanged callbacks for all items
        local backpackSize = self.backpack:GetSize();
        for j = 1, backpackSize, 1 do
            local item = self.backpack:GetItem(j);
            if item and item.QuantityChanged == nil then
                item.QuantityChanged = function(sender, args)
                    if item:GetCategory() == 164 then
                        mainWindow:UpdateCount(item:GetName());
                    end
                end
            end
        end
    end

    -- Refresh list on backpack item added event
    self.backpack.ItemAdded = function(sender, args)
        local item = self.backpack:GetItem(args.Index);
        if item:GetCategory() == 164 then
            mainWindow:UpdateCount(item:GetName());
        end
        if item.QuantityChanged == nil then
            item.QuantityChanged = function(sender, args)
                if item:GetCategory() == 164 then
                    mainWindow:UpdateCount(item:GetName());
                end
            end
        end
    end

    -- Refresh list on backpack item removed event
    self.backpack.ItemRemoved = function(sender, args)
        local item = self.backpack:GetItem(args.Index);
        if item:GetCategory() == 164 then
            mainWindow:UpdateCount(item:GetName(), item:GetQuantity());
        end
    end

    -- Create the trees
    self.trees = { };
    self.nOfTrees = 0;
    for i, item in ipairs(BusyBeeTreeData) do
        self.nOfTrees = self.nOfTrees + 1;
        self.trees[self.nOfTrees] = BusyBeeTree(item.data, item.fields);
        self.trees[self.nOfTrees]:SetParent(self);
        self.trees[self.nOfTrees]:SetVisible(self.settings.visibleTree == self.nOfTrees);
        if self.settings.treeStates ~= nil and self.settings.treeStates[self.nOfTrees] then
            self.trees[self.nOfTrees]:RestoreState(self.settings.treeStates[self.nOfTrees]);
        end
    end
    
    -- Create the item count label
    self.itemLabel = Turbine.UI.Label();
    self.itemLabel:SetParent(self);
    self.itemLabel:SetPosition(10, 10);
    self.itemLabel:SetSize(200, 20);
    self.itemLabel:SetFontStyle(Turbine.UI.FontStyle.Outline);
    self.itemLabel:SetForeColor(ItemLabelColor);
    self.itemLabel:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
    self.itemLabel:SetFont(Turbine.UI.Lotro.Font.TrajanPro13);
    self.itemLabel:SetText("");
    
    self:UpdateAllTasks();
    self:UpdateCount();
    self:PerformLayout();
end

--------------------------------------------------------------------------------
function BusyBeeWindow:SetPassiveOpacity(opacity)
    windowOpacityPassive = opacity;
    self:SetOpacity(windowOpacityPassive);
end

--------------------------------------------------------------------------------
function BusyBeeWindow:UpdateVisibility(item, fForceUpdate)

    local player = Turbine.Gameplay.LocalPlayer.GetInstance();
    local playerLevel = player:GetLevel();
    local fChanged = false;
    
    if item == nil then
        for i, item2 in ipairs(BusyBeeTasks) do
            local fChanged2 = self:UpdateVisibility(item2, true);
            fChanged = fChanged or fChanged2;
        end
    elseif item.id == nil then
        for i, id in ipairs(item.ids) do
            local fChanged2 = self:UpdateVisibility(BusyBeeTasks[id]);
            fChanged = fChanged or fChanged2;
        end
    else
        local isVisible = true;
        local fullname = BusyBeeItemName(item);
        if not self.fShowZero and BusyBeeCounts[fullname].count == 0 then
            isVisible = false;
        elseif not self.fShowAllLevels and (playerLevel > item.lvlmax or playerLevel < item.lvlmin) then
            isVisible = false;
        elseif self.fShowRepAward and item.faction == "(none)" then
            isVisible = false;
        end
        
        if fForceUpdate or isVisible ~= item.visible then
            fChanged = true;
            item.visible = isVisible;
            for i, node in ipairs(item.nodes) do
                node:FixVisibility();
            end
        end
    end
    
    return fChanged;
end

--------------------------------------------------------------------------------
function BusyBeeWindow:UpdateCount(itemName, delta)

    local backpackSize = self.backpack:GetSize();
    local fVisibilityChanged = false;
    local player = Turbine.Gameplay.LocalPlayer.GetInstance();
    local playerLevel = player:GetLevel();
    
    -- Recalculate backpack contents
    local counts = { };
    for i = 1, backpackSize, 1 do
        local item = self.backpack:GetItem(i);
        if item ~= nil and BusyBeeCounts[item:GetName()] ~= nil then
            if counts[item:GetName()] == nil then
                if itemName == item:GetName() and delta then
                    counts[item:GetName()] = item:GetQuantity() - delta;
                else
                    counts[item:GetName()] = item:GetQuantity();
                end
            else
                counts[item:GetName()] = counts[item:GetName()] + item:GetQuantity();
            end
        end
    end
    
    for key, value in pairs(BusyBeeCounts) do
        if itemName == nil or itemName == key then
            local count = 0;
            if counts[key] ~= nil then
                count = counts[key];
            end
            
            if BusyBeeCounts[key].count ~= count then
                
                if BusyBeeCounts[key].count < BusyBeeCounts[key].quantity and count >= BusyBeeCounts[key].quantity then
                    for i, id in ipairs(BusyBeeCounts[key].ids) do
                        if (BusyBeeTasks[id].lvlmin <= playerLevel and BusyBeeTasks[id].lvlmax >= playerLevel) then
                            Turbine.Shell.WriteLine(string.format("%s(%d) can be turned in at %s for %s.",
                                key, BusyBeeCounts[key].quantity, BusyBeeTasks[id].location, BusyBeeTasks[id].faction));
                        end
                    end
                end
                BusyBeeCounts[key].count = count;
                
                for i, node in ipairs(BusyBeeCounts[key].nodes) do
                    node:UpdateCount();
                end
                
                local fVisibilityChanged2 = self:UpdateVisibility(BusyBeeCounts[key]);
                fVisibilityChanged = fVisibilityChanged or fVisibilityChanged2;
                
            end
        end
    end
    
    self:UpdateItemLabel();
end

--------------------------------------------------------------------------------
function BusyBeeWindow:UpdateItemLabel()
    local nVisible = 0;
    for key, value in pairs(BusyBeeCounts) do
        local fVisible = 0;
        for i, id in ipairs(value.ids) do
            if BusyBeeTasks[id].visible then
                fVisible = 1;
                break;
            end
        end
        nVisible = nVisible + fVisible;
    end
    if nVisible == 0 then
        self.itemLabel:SetText("");
    elseif nVisible == 1 then
        self.itemLabel:SetText(nVisible .. " busy bee item.");
    else
        self.itemLabel:SetText(nVisible .. " busy bee items.");
    end
end

--------------------------------------------------------------------------------
function BusyBeeWindow:UpdateAllTasks()
    local player = Turbine.Gameplay.LocalPlayer.GetInstance();
    local playerLevel = player:GetLevel();
    for i, item in pairs(BusyBeeTasks) do
        local isVisible = true;
        if not self.fShowZero and BusyBeeCounts[item.fullname].count == 0 then
            isVisible = false;
        elseif not self.fShowAllLevels and (playerLevel > item.lvlmax or playerLevel < item.lvlmin) then
            isVisible = false;
        elseif self.fShowRepAward and item.faction == "(none)" then
            isVisible = false;
        end
        
        if isVisible ~= item.visible then
            item.visible = isVisible;
            for i, node in ipairs(item.nodes) do
                node:FixVisibility();
            end
        end
    end
    
    self:UpdateItemLabel();
end

--------------------------------------------------------------------------------
function BusyBeeWindow:PerformLayout()
    self:Layout({ })
end

--------------------------------------------------------------------------------
function BusyBeeWindow:Layout(args)
    local width, height = self:GetSize();
    
    -- Position the buttons
    if self.fShowButtons then
        local buttonLeft = marginX;
        self.ModeButton:SetPosition(buttonLeft, marginTop);
        self.ModeButton:SetSize(buttonWidth1 - 5, buttonHeight);
        buttonLeft = buttonLeft + buttonWidth1;
        self.ItemCountButton:SetPosition(buttonLeft, marginTop);
        self.ItemCountButton:SetSize(buttonWidth1, buttonHeight);
        buttonLeft = buttonLeft + buttonWidth1;
        self.AllLevelsButton:SetPosition(buttonLeft, marginTop);
        self.AllLevelsButton:SetSize(buttonWidth1, buttonHeight);
        buttonLeft = buttonLeft + buttonWidth1;
        self.RepAwardButton:SetPosition(buttonLeft, marginTop);
        self.RepAwardButton:SetSize(buttonWidth1, buttonHeight);
        buttonLeft = buttonLeft + buttonWidth1;
        
        if (buttonLeft < width - marginX - 2 * buttonWidth2) then
            buttonLeft = width - marginX - 2 * buttonWidth2;
        end
        
        self.ExpandButton:SetPosition(buttonLeft, marginTop);
        self.ExpandButton:SetSize(buttonWidth2, buttonHeight);
        buttonLeft = buttonLeft + buttonWidth2;
        self.CollapseButton:SetPosition(buttonLeft, marginTop);
        self.CollapseButton:SetSize(buttonWidth2, buttonHeight);
        
        -- Set button text
        if self.fShowZero then
            self.ItemCountButton:SetText("All Cnt");
        else
            self.ItemCountButton:SetText("Cnt > 0");
        end
        
        if self.fShowAllLevels then
            self.AllLevelsButton:SetText("All Lvl");
        else
            self.AllLevelsButton:SetText("My Lvl");
        end
        
        if self.fShowRepAward then
            self.RepAwardButton:SetText("Rep > 0");
        else
            self.RepAwardButton:SetText("All Rep");
        end
    end
    
    -- Show or hide the buttons
    self.ModeButton:SetVisible(self.fShowButtons);
    self.ItemCountButton:SetVisible(self.fShowButtons);
    self.AllLevelsButton:SetVisible(self.fShowButtons);
    self.RepAwardButton:SetVisible(self.fShowButtons);
    self.ExpandButton:SetVisible(self.fShowButtons);
    self.CollapseButton:SetVisible(self.fShowButtons);
    
    -- Position the tree views
    local listTop = marginTop + buttonHeight;
    local listWidth = width - 2 * marginX;
    local listHeight = height - listTop - marginBottom;
    for i, tree in ipairs(self.trees) do
        tree:SetPosition(marginX, listTop);
        tree:SetSize(listWidth, listHeight);
    end
    
    self.itemLabel:SetPosition(marginX, marginTop);
    self.itemLabel:SetSize(listWidth, buttonHeight);
    self.itemLabel:SetVisible(not self.fShowButtons);
end

--------------------------------------------------------------------------------
function BusyBeeWindow:LoadSettings()
    self.settings = Turbine.PluginData.Load(Turbine.DataScope.Character, "BusyBeeSettings");
    if (type(self.settings) ~= "table"  or self.settings.version ~= BusyBeeSettingsVersion) then
        self.settings = { };
        self.settings.version = BusyBeeSettingsVersion;
        self.settings.windowTop = 0;
        self.settings.windowLeft = (Turbine.UI.Display.GetWidth() - windowOriginalWidth) / 2;
        self.settings.windowWidth = windowOriginalWidth;
        self.settings.windowHeight = windowOriginalHeight;
        self.settings.passiveOpacity = windowOpacityPassive;
        self.settings.treeStates = { };
        self.settings.visibleTree = 1;
        self.settings.fShowZero = true;
        self.settings.fShowAllLevels = false;
        self.settings.fShowRepAward = false;
        Turbine.Shell.WriteLine("Busy Bee Tasks initialized.");
    else
        windowOpacityPassive = self.settings.passiveOpacity;
        Turbine.Shell.WriteLine("Busy Bee Tasks loaded.");
    end
    
    if self.settings.windowLeft + self.settings.windowWidth > Turbine.UI.Display.GetWidth() then
        self.settings.windowLeft = Turbine.UI.Display.GetWidth() - self.settings.windowWidth;
    end
    if self.settings.windowTop + self.settings.windowHeight > Turbine.UI.Display.GetHeight() then
        self.settings.windowTop = Turbine.UI.Display.GetHeight() - self.settings.windowHeight;
    end
end

--------------------------------------------------------------------------------
function BusyBeeWindow:SaveSettings()
    local width, height = self:GetSize();
    local left, top = self:GetPosition();
    
    local visibleTree = 1;
    treeStates = { };
    for i, tree in ipairs(self.trees) do
        if tree:IsVisible() then
            visibleTree = i;
        end
        treeStates[i] = tree:GetState();
    end

    if (self.settings.windowTop ~= top or
        self.settings.windowLeft ~= left or
        self.settings.windowWidth ~= width or
        self.settings.windowHeight ~= height or
        self.settings.passiveOpacity ~= windowOpacityPassive or
        self.settings.treeStates ~= treeStates or
        self.settings.visibleTree ~= visibleTree or
        self.settings.fShowZero ~= self.fShowZero or
        self.settings.fShowAllLevels ~= self.fShowAllLevels or
        self.settings.fShowRepAward ~= self.fShowRepAward) then
        
        self.settings.windowTop = top;
        self.settings.windowLeft = left;
        self.settings.windowWidth = width;
        self.settings.windowHeight = height;
        self.settings.passiveOpacity = windowOpacityPassive;
        self.settings.treeStates = treeStates;
        self.settings.visibleTree = visibleTree;
        self.settings.fShowZero = self.fShowZero;
        self.settings.fShowAllLevels = self.fShowAllLevels;
        self.settings.fShowRepAward = self.fShowRepAward;
        
        Turbine.PluginData.Save(Turbine.DataScope.Character, "BusyBeeSettings", self.settings);
--         Turbine.Shell.WriteLine("Busy Bee Tasks saved.");
    end
end
